#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that implements winreg for Windows Subsystem for Linux
"""

## \package wslwinreg.wslapi

# Disable camel case requirement for function names
# pylint: disable=invalid-name
# pylint: disable=useless-object-inheritance
# pylint: disable=unused-argument
# pylint: disable=broad-except
# pylint: disable=redefined-builtin
# pylint: disable=raise-missing-from

import os
import stat
import subprocess
import socket
import platform
import struct
from enum import IntEnum
from .common import KEY_WRITE, KEY_WOW64_64KEY, KEY_READ, PY2, \
    winerror_to_errno, builtins, ERROR_FILE_NOT_FOUND, from_registry_bytes, \
    REG_SZ, to_registry_bytes


## Type long for Python 2 compatibility
try:
    long
except NameError:
    # Fake it for Python 3
    long = int

## Type basestring for Python 2 compatibility
try:
    basestring
except NameError:
    # Fake it for Python 3
    basestring = str

class Commands(IntEnum):
    """
    Commands to send to the bridging executable.
    """

    ## Exit the server
    ABORT = 0

    ## Connection handshake
    CONNECT = 1

    ## Perform CloseKey()
    CLOSE_KEY = 2

    ## Perform ConnectRegistry()
    CONNECT_REGISTRY = 3

    ## Perform CreateKey()
    CREATE_KEY = 4

    ## Perform CreateKeyEx()
    CREATE_KEY_EX = 5

    ## Perform DeleteKey()
    DELETE_KEY = 6

    ## Perform DeleteKeyEx()
    DELETE_KEY_EX = 7

    ## Perform DeleteValue()
    DELETE_VALUE = 8

    ## Perform EnumKey()
    ENUM_KEY = 9

    ## Perform EnumValue()
    ENUM_VALUE = 10

    ## Perform ExpandEnvironmentStrings()
    EXPAND_ENVIRONMENTSTRINGS = 11

    ## Perform FlushKey()
    FLUSH_KEY = 12

    ## Perform LoadKey()
    LOAD_KEY = 13

    ## Perform OpenKey()
    OPEN_KEY = 14

    ## Perform OpenKeyEx()
    OPEN_KEY_EX = 15

    ## Perform QueryInfoKey()
    QUERY_INFO_KEY = 16

    ## Perform QueryValue()
    QUERY_VALUE = 17

    ## Perform QueryValueEx()
    QUERY_VALUE_EX = 18

    ## Perform SaveKey()
    SAVE_KEY = 19

    ## Perform SetValue()
    SET_VALUE = 20

    ## Perform SetValueEx()
    SET_VALUE_EX = 21

    ## Perform DisableReflectionKey()
    DISABLE_REFLECTION_KEY = 22

    ## Perform EnableReflectionKey()
    ENABLE_REFLECTION_KEY = 23

    ## Perform QueryReflectionKey()
    QUERY_REFLECTION_KEY = 24

    ## Perform get_file_into()
    GET_FILE_INFO = 25

## Loopback address
_LOCALHOST = '127.0.0.1'

## Transmission buffer size
_BUFFER_SIZE = 1024

## Directory for the windows executables
_WIN_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'bin')

## Set the exe suffix for the CPU in use
_EXESUFFIX = machine = platform.machine().lower()

# Use a common suffix for amd64 instruction sets
if _EXESUFFIX in ('amd64', 'x86_64', 'em64t'):
    _EXESUFFIX = 'x64'

## Patch to the executable to bridge
_WIN_EXE = os.path.join(_WIN_DIR, 'backend-' + _EXESUFFIX + '.exe')

# Make sure it's executable

## Saved stat for the backend server
_STAT = os.stat(_WIN_EXE)
if not _STAT.st_mode & stat.S_IEXEC:
    os.chmod(_WIN_EXE, _STAT.st_mode | stat.S_IEXEC)
del _STAT

# Prepare a socket to be waiting for the exe once it is launched

## Socket used for listening for the exe
_LISTEN_SOCKET = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
_LISTEN_SOCKET.bind((_LOCALHOST, 0))

## Semi-random port assigned to the socket by the operating system
_LISTEN_PORT = _LISTEN_SOCKET.getsockname()[1]

# Start listening
_LISTEN_SOCKET.listen(1)

try:
    ## Popen object for the bridge executable
    _EXEC_FP = subprocess.Popen(
        (_WIN_EXE, '-p', str(_LISTEN_PORT)),
        cwd=_WIN_DIR, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        universal_newlines=True)
except OSError:
    raise ImportError("Windows executable {} for bridging not found.".format(
        _WIN_EXE))

# At thie point, the exe had started, connect to it.
_LISTEN_SOCKET.settimeout(3.0)
try:
    ## @var _CONNECTION_ADDR
    # Connection address

    ## Connection socket
    _CONNECTION_SOCKET, _CONNECTION_ADDR = _LISTEN_SOCKET.accept()
except socket.timeout:
    raise ImportError("Failure to connect with bridging executable")

# Set the timeout
_CONNECTION_SOCKET.settimeout(5.0)

if _CONNECTION_SOCKET.recv(_BUFFER_SIZE) != b"Bridge started 1.0":
    raise ImportError("Windows Bridge version mismatch")

########################################


class WindowsError(OSError):
    """
    This exception doesn't exist in Cygwin/MSYS, provide it.
    """

    # pylint: disable=super-init-not-called
    def __init__(self, winerror, strerror=None, filename=None):
        """
        Initialize a WindowsError exception

        Args:
            winerror: The windows error code
            strerror: The string describing the error
            filename: Name of the file that caused this error, if applicable.
        """

        ## Windows error code
        self.winerror = winerror

        ## Linux style error code
        self.errno = winerror_to_errno(winerror)

        ## Description string of the error
        self.strerror = strerror

        ## Name of the responsible file for this error
        self.filename = filename

    def __str__(self):
        """
        Convert the class into a string.

        Returns:
            String describing the error contained.
        """
        if self.filename is not None:
            return "[Error %s] %s: %s" % (self.winerror, self.strerror,
                                          repr(self.filename))
        return "[Error %s] %s" % (self.winerror, self.strerror)


# Ensure this exception is available globally.
builtins.WindowsError = WindowsError

########################################


def handleLRESULT():
    """
    Receive the LRESULT from the bridge.
    """

    # Get the LRESULT
    data = _CONNECTION_SOCKET.recv(4)
    # Error code
    return_code = struct.unpack('<I', data)[0]
    if return_code:

        # Parse out the utf-8 error string
        error_str = recv_string()
        # Special case, unit tests require a FileNotFoundError
        if return_code == ERROR_FILE_NOT_FOUND:
            raise FileNotFoundError(error_str)

        # Throw a WindowsError exception
        raise WindowsError(return_code, error_str)

########################################


def create_string_buffer(temp_string, is_binary=False):
    """
    Convert a string into a utf-8 byte stream

    Args:
        temp_string: String to convert
        is_binary: True if the input in binary, not a string.
    """

    # Ensure it's a string
    if temp_string:
        if not is_binary:
            temp_string = temp_string.encode('utf-8')
    else:
        temp_string = b''

    temp_string_length = len(temp_string)

    # Get the length
    buffer = struct.pack(
        '<I',
        temp_string_length)

    if not temp_string_length:
        return buffer

    # Append the string to the buffer and send the length and string
    return buffer + temp_string

########################################


def recv_string(convert_to_string=True):
    """
    Recieve a string from the socket

    Returns:
        UTF-8 string.
    """

    # Get the result string length
    data = _CONNECTION_SOCKET.recv(4)
    new_string_length = struct.unpack('<I', data)[0]

    # Get the result string
    if new_string_length:

        # Create the result buffer
        data = bytearray()
        # All data received?
        while len(data) < new_string_length:
            # Get a chunk
            packet = _CONNECTION_SOCKET.recv(new_string_length - len(data))
            if not packet:
                raise socket.timeout("Connection broken")
            data.extend(packet)
            del packet
        data = bytes(data)
    else:
        data = b''

    if convert_to_string:
        data = data.decode('utf-8')
    return data

########################################


def _RegCloseKey(hkey):
    """
    Low level function to call RegCloseKey
    """

    # Send the command and a QWORD of the pointer
    buffer = struct.pack(
        '<BQ', Commands.CLOSE_KEY.value,
        hkey)
    _CONNECTION_SOCKET.sendall(buffer)

    handleLRESULT()

########################################


class PyHKEY(object):
    """
    A Python object representing a win32 registry key.

    This object wraps a Windows HKEY object, automatically closing it when the
    object is destroyed. To guarantee cleanup, you can call either the Close()
    method on the object, or the CloseKey() function.

    All registry functions in this module return one of these objects.

    All registry functions in this module which accept a handle object also
    accept an integer, however, use of the handle object is encouraged.

    Handle objects provide semantics for __bool__() – thus

    @code
        if handle:
            print("Yes")
    @endcode

    will print Yes if the handle is currently valid (has not been closed or
    detached).

    The object also support comparison semantics, so handle objects will compare
    true if they both reference the same underlying Windows handle value.

    Handle objects can be converted to an integer (e.g., using the built-in
    int() function), in which case the underlying Windows handle value is
    returned. You can also use the Detach() method to return the integer handle,
    and also disconnect the Windows handle from the handle object.
    """

    def __init__(self, hkey, null_ok=False):
        """
        Initialize the PyHKEY class.

        Args:
            hkey: Integer value representing the pointer to the HKEY
            null_ok: True if None is acceptable.
        """
        if not isinstance(hkey, (int, long)):
            if not null_ok or not isinstance(hkey, type(None)):
                raise TypeError("A handle must be an integer")

        ## Integer that represents the HANDLE pointer
        self.hkey = hkey

    def __del__(self):
        """
        Called when this object is garbage collected.
        """
        if self.hkey:
            # Ignore errors.
            try:
                self.Close()
            except BaseException:
                pass

    def Close(self):
        """
        Closes the underlying Windows handle.

        Note:
            If the handle is already closed, no error is raised.
        """
        if self.hkey:
            _RegCloseKey(self.hkey)
        self.hkey = 0

    def Detach(self):
        """
        Detaches the Windows handle from the handle object.

        Note:
        The result is the value of the handle before it is detached. If the
        handle is already detached, this will return zero.

        After calling this function, the handle is effectively invalidated,
        but the handle is not closed.  You would call this function when you
        need the underlying win32 handle to exist beyond the lifetime of the
        handle object.

        On 64 bit windows, the result of this function is a long integer.

        Returns:
            Previous HKEY integer.
        """
        hkey = self.hkey
        self.hkey = 0
        return hkey

    def _handle(self):
        """
        The integer Win32 handle.
        """
        return self.hkey

    # The integer Win32 handle.
    handle = property(_handle)

    def __enter__(self):
        """
        Called when object is entered.

        Note:
            Needed for the Python ``with`` statement.
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Release handle on class destruction.

        Note:
            Needed for the Python ``with`` statement.
        """
        self.Close()
        return False

    def __hash__(self):
        """
        Convert the object into a hash.

        Note:
            The implementation returns the id, which should be
            random enough.
        """
        return id(self)

    def __int__(self):
        """
        Converting a handle to an integer returns the Win32 handle.
        """
        return self.hkey

    if PY2:
        def __nonzero__(self):
            """
            Handles with an open object return true, otherwise false.
            """
            return bool(self.hkey)
    else:
        def __bool__(self):
            """
            Handles with an open object return true, otherwise false.
            """
            return bool(self.hkey)

    def __repr__(self):
        """
        Return descriptive string for the class object.
        """
        return "<PyHKEY at %08X (%08X)>" % (id(self), self.hkey)

    def __str__(self):
        """
        Return short string for the handle object.
        """
        return "<PyHKEY:%08X>" % self.hkey

    # Python sort description.
    _as_parameter_ = property(lambda self: self.hkey)

    @staticmethod
    def make(hkey, null_ok=None):
        """
        Convert a pointer into a PyHKEY object.
        """

        # Already converted?
        if isinstance(hkey, PyHKEY):
            return hkey

        # Is it a valid integer?
        if isinstance(hkey, (int, long)):
            return PyHKEY(hkey)

        # None, if None is allowed.
        if null_ok and isinstance(hkey, type(None)):
            return PyHKEY(None)

        # This didn't end well.
        raise TypeError("A handle must be a HKEY object or an integer")

########################################


def test_string(input_string):
    """
    Raise an exception if the input is not None or a string.

    Args:
        input_string: String to test.
    Exception:
        ``TypeError``
    """

    if input_string is None:
        return

    if not isinstance(input_string, basestring):
        raise TypeError("{} has to be None or a string".format(input_string))

########################################


def CloseKey(hkey):
    """
    Closes a previously opened registry key.

    The hkey argument specifies a previously opened key.

    Args:
        hkey: PyHKEY object or None.

    Note:
        If hkey is not closed using this method (or via hkey.Close()),
        it is closed when the hkey object is destroyed by Python.

    """
    PyHKEY.make(hkey).Close()

########################################


def ConnectRegistry(computer_name, key):
    """
    Establishes a connection to a predefined registry handle.

    Establishes a connection to a predefined registry handle on another
    computer, and returns a handle object.

    Args:
        computer_name: Is the name of the remote computer, of the form
            ``r"\\computername"``. If ``None``, the local computer is used.
        key: Is the predefined handle to connect to.
    Returns:
        PyHKEY object
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    test_string(computer_name)
    buffer = struct.pack(
        '<BQ',
        Commands.CONNECT_REGISTRY.value,
        PyHKEY.make(key).hkey)

    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(computer_name))

    # This function CAN take a while, so allow it to take the time
    _CONNECTION_SOCKET.settimeout(20.0)
    data = _CONNECTION_SOCKET.recv(8)

    # Restore the proper timeout
    _CONNECTION_SOCKET.settimeout(3.0)
    handleLRESULT()
    return PyHKEY.make(struct.unpack('<Q', data)[0])

########################################


def CreateKey(key, sub_key):
    """
    Creates or opens the specified key.

    If key is one of the predefined keys, sub_key may be None. In that case,
    the handle returned is the same key handle passed in to the function.

    If the key already exists, this function opens the existing key.

    Args:
        key: Is an already open key, or one of the predefined HKEY_* constants.
        sub_key: Is a string that names the key this method opens or creates.
    Returns:
        Handle of the opened key.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    test_string(sub_key)

    buffer = struct.pack(
        '<BQ',
        Commands.CREATE_KEY.value,
        PyHKEY.make(key).hkey)

    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(sub_key))

    data = _CONNECTION_SOCKET.recv(8)
    handleLRESULT()
    return PyHKEY.make(struct.unpack('<Q', data)[0])

########################################


def CreateKeyEx(key, sub_key, reserved=0, access=KEY_WRITE):
    """
    Creates or opens the specified key.

    If key is one of the predefined keys, sub_key may be None. In that case,
    the handle returned is the same key handle passed in to the function.

    If the key already exists, this function opens the existing key.

    Args:
        key: Is an already open key, or one of the predefined HKEY_* constants
        sub_key: Is a string that names the key this method opens or creates.
        reserved: Is a reserved integer, and must be zero. The default is zero.
        access: Os an integer that specifies an access mask that describes the
            desired security access for the key. Default is common.KEY_WRITE.

    Returns:
        Handle of the opened key.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    test_string(sub_key)

    buffer = struct.pack(
        '<BQII',
        Commands.CREATE_KEY_EX.value,
        PyHKEY.make(key).hkey, reserved, access)

    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(sub_key))

    data = _CONNECTION_SOCKET.recv(8)
    handleLRESULT()
    return PyHKEY.make(struct.unpack('<Q', data)[0])

########################################


def DeleteKey(key, sub_key):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    test_string(sub_key)

    buffer = struct.pack(
        '<BQ',
        Commands.DELETE_KEY.value,
        PyHKEY.make(key).hkey)

    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(sub_key))
    handleLRESULT()

########################################


def DeleteKeyEx(key, sub_key, access=KEY_WOW64_64KEY, reserved=0):
    """
    Deletes the specified key.

    If the method succeeds, the entire key, including all of its values,
    is removed.

    Note:
        This method can not delete keys with subkeys.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.

        sub_key: Os a string that must be a subkey of the key identified by the
            key parameter. This value must not be ``None``, and the key may not
            have subkeys.

        access: Is an integer that specifies an access mask that describes the
            desired security access for the key. Default
            is common.KEY_WOW64_64KEY.

        reserved: Is a reserved integer, and must be zero. The default is zero.
    Exception:
        ``WindowsError``
    """

    test_string(sub_key)

    buffer = struct.pack(
        '<BQII',
        Commands.DELETE_KEY_EX.value,
        PyHKEY.make(key).hkey, reserved, access)

    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(sub_key))
    handleLRESULT()

########################################


def DeleteValue(key, value):
    """
    Removes a named value from a registry key.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        value: Is a string that identifies the value to remove.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    test_string(value)

    buffer = struct.pack(
        '<BQ',
        Commands.DELETE_VALUE.value,
        PyHKEY.make(key).hkey)

    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(value))
    handleLRESULT()

########################################


def EnumKey(key, index):
    """
    Enumerates subkeys of an open registry key, returning a string.

    The function retrieves the name of one subkey each time it is called.
    It is typically called repeatedly until an ``OSError`` exception is
    raised, indicating no more values are available.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        index: Is an integer that identifies the index of the key to retrieve.

    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    buffer = struct.pack(
        '<BQI',
        Commands.ENUM_KEY.value,
        PyHKEY.make(key).hkey, index)

    _CONNECTION_SOCKET.sendall(buffer)

    new_data = recv_string()
    # Handle the error
    handleLRESULT()
    return new_data

########################################


def EnumValue(key, index):
    # pylint: disable=line-too-long
    """
    Enumerates values of an open registry key, returning a tuple.

    The function retrieves the name of one subkey each time it is called.
    It is typically called repeatedly, until an ``OSError`` exception is
    raised, indicating no more values.

    | Index | Meaning |
    | ----- | ------- |
    | 0 | A string that identifies the value. |
    | 1 | An object that holds the value data, and whose type depends on the underlying registry type |
    | 2 | An integer that identifies the type of the value data |

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        index: Is an integer that identifies the index of the value to retrieve.
    Returns:
       A tuple of 3 items.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    buffer = struct.pack(
        '<BQI',
        Commands.ENUM_VALUE.value,
        PyHKEY.make(key).hkey, index)

    _CONNECTION_SOCKET.sendall(buffer)

    # Get the id string, data and the type
    new_string = recv_string()
    new_data = recv_string(convert_to_string=False)
    data = _CONNECTION_SOCKET.recv(4)

    # Handle the error
    handleLRESULT()
    typ = struct.unpack('<I', data)[0]
    return (new_string,
            from_registry_bytes(new_data, len(new_data), typ),
            typ)

########################################


def ExpandEnvironmentStrings(str):
    """
    Expands environment variables.

    Expands environment variable placeholders %NAME% in strings
    like REG_EXPAND_SZ.

    Args:
        str: String to expand.
    Returns:
        Expanded string
    Exception:
        ``WindowsError``
    """

    test_string(str)

    # Send the command
    buffer = struct.pack(
        '<B',
        Commands.EXPAND_ENVIRONMENTSTRINGS.value)

    # Append the string to the buffer
    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(str))

    # Get the answer
    new_string = recv_string()

    # Get the error
    handleLRESULT()
    return new_string

########################################


def FlushKey(key):
    """
    Writes all the attributes of a key to the registry.

    It is not necessary to call ``FlushKey()`` to change a key. Registry
    changes are flushed to disk by the registry using its lazy flusher.
    Registry changes are also flushed to disk at system shutdown. Unlike
    ``CloseKey()``, the ``FlushKey()`` method returns only when all the
    data has been written to the registry. An application should only
    call ``FlushKey()`` if it requires absolute certainty that registry
    changes are on disk.

    Note:
        If you don’t know whether a ``FlushKey()`` call is required, it
        probably isn’t.
    Args:
        key: Is an already open key, or any one of the predefined
        HKEY_* constants.
    Exception:
        ``WindowsError``
    """

    # Send the command and a QWORD of the pointer
    buffer = struct.pack(
        '<BQ',
        Commands.FLUSH_KEY.value,
        PyHKEY.make(key).hkey)
    _CONNECTION_SOCKET.sendall(buffer)

    handleLRESULT()

########################################


def LoadKey(key, sub_key, file_name):
    """
    Creates a subkey under the specified key.

    Creates a subkey under the specified key and stores registration
    information from a specified file into that subkey.

    This file must have been created with the SaveKey() function.
    Under the file allocation table (FAT) file system, the filename may not
    have an extension.

    A call to ``LoadKey()`` fails if the calling process does not have the
    SE_RESTORE_PRIVILEGE privilege.

    If key is a handle returned by ``ConnectRegistry()``, then the path
    specified in fileName is relative to the remote computer.

    Args:
        key: Is a handle returned by ``ConnectRegistry()`` or one of the
            constants common.HKEY_USERS or common.HKEY_LOCAL_MACHINE.
        sub_key: Is a string that identifies the sub_key to load
        file_name: Is the name of the file to load registry data from.

    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    test_string(sub_key)
    test_string(file_name)

    test_string(sub_key)
    buffer = struct.pack(
        '<BQ',
        Commands.LOAD_KEY.value,
        PyHKEY.make(key).hkey)

    _CONNECTION_SOCKET.sendall(
        buffer +
        create_string_buffer(sub_key) +
        create_string_buffer(file_name))
    handleLRESULT()

########################################


def OpenKey(key, sub_key, reserved=0, access=KEY_READ):
    """
    Opens the specified key, returning a handle object.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        sub_key: Is a string that identifies the sub_key to open
        reserved: Is a reserved integer, and must be zero.
            Default is zero.
        access: Is an integer that specifies an access mask that
            describes the desired security access for the key. Default
            is KEY_READ.

    Returns:
        A new handle to the specified key.

    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    test_string(sub_key)
    buffer = struct.pack(
        '<BQII',
        Commands.OPEN_KEY.value,
        PyHKEY.make(key).hkey,
        reserved,
        access)

    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(sub_key))

    data = _CONNECTION_SOCKET.recv(8)
    handleLRESULT()
    return PyHKEY.make(struct.unpack('<Q', data)[0])

########################################


def OpenKeyEx(key, sub_key, reserved=0, access=KEY_READ):
    """
    Opens the specified key, returning a handle object.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        sub_key: Is a string that identifies the sub_key to open
        reserved: Is a reserved integer, and must be zero.
            Default is zero.
        access: Is an integer that specifies an access mask that
            describes the desired security access for the key. Default
            is KEY_READ.

    Returns:
        A new handle to the specified key.

    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """
    return OpenKey(key, sub_key, reserved, access)

########################################


def QueryInfoKey(key):
    # pylint: disable=line-too-long
    """
    Returns information about a key, as a tuple.

    | Index | Meaning |
    | ----- | ------- |
    | 0 | An integer giving the number of sub keys this key has. |
    | 1 | An integer giving the number of values this key has. |
    | 2 | An integer giving when the key was last modified (if available) as 100’s of nanoseconds since Jan 1, 1601. |

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
    Returns:
        A tuple of 3 items.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    # Send the command and a QWORD of the pointer
    buffer = struct.pack(
        '<BQ',
        Commands.QUERY_INFO_KEY.value,
        PyHKEY.make(key).hkey)
    _CONNECTION_SOCKET.sendall(buffer)

    data = _CONNECTION_SOCKET.recv(4 + 4 + 8)
    handleLRESULT()
    return struct.unpack('<IIQ', data)

########################################


def QueryValue(key, sub_key):
    """
    Retrieves the unnamed value for a key, as a string.

    Values in the registry have name, type, and data components. This method
    retrieves the data for a key’s first value that has a NULL name. But the
    underlying API call doesn’t return the type, so always use
    ``QueryValueEx()`` if possible.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        sub_key: Is a string that holds the name of the subkey with which the
            value is associated. If this parameter is None or empty, the
            function retrieves the value set by the ``SetValue()`` method for
            the key identified by ``key``.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    test_string(sub_key)
    buffer = struct.pack(
        '<BQ',
        Commands.QUERY_VALUE.value,
        PyHKEY.make(key).hkey)

    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(sub_key))

    new_data = recv_string()
    handleLRESULT()
    return new_data

########################################


def QueryValueEx(key, value_name):
    """
    Retrieves the type and data for a specified value name.

    Retrieves the type and data for a specified value name associated with an
    open registry key.

    | Index | Meaning |
    | ----- | ------- |
    | 0 | The value of the registry item. |
    | 1 | An integer giving the registry type for this value. |

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        value_name: Is a string indicating the value to query.
    Returns:
        A tuple of 2 items.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    test_string(value_name)
    buffer = struct.pack(
        '<BQ',
        Commands.QUERY_VALUE_EX.value,
        PyHKEY.make(key).hkey)

    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(value_name))

    new_data = recv_string(convert_to_string=False)
    data = _CONNECTION_SOCKET.recv(4)

    # Restore the proper timeout
    handleLRESULT()
    typ = struct.unpack('<I', data)[0]
    return (from_registry_bytes(new_data, len(new_data), typ),
            typ)

########################################


def SaveKey(key, file_name):
    """
    Saves the specified key, and all its subkeys to the specified file.

    If key represents a key on a remote computer, the path described by
    file_name is relative to the remote computer. The caller of this method
    must possess the SeBackupPrivilege security privilege.

    This function passes NULL for security_attributes to the API.

    Note:
        Privileges are different than permissions.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        file_name: Is the name of the file to save registry data to.
            This file cannot already exist. If this filename includes an
            extension, it cannot be used on file allocation table (FAT) file
            systems by the ``LoadKey()`` method.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    test_string(file_name)
    buffer = struct.pack(
        '<BQ',
        Commands.SAVE_KEY.value,
        PyHKEY.make(key).hkey)

    _CONNECTION_SOCKET.sendall(buffer + create_string_buffer(file_name))

    handleLRESULT()

########################################


def SetValue(key, sub_key, type, value):
    """
    Associates a value with a specified key.

    If the key specified by the sub_key parameter does not exist,
    the SetValue function creates it.

    Value lengths are limited by available memory. Long values
    (more than 2048 bytes) should be stored as files with the
    filenames stored in the configuration registry.
    This helps the registry perform efficiently.

    The key identified by the ``key`` parameter must have been
    opened with common.KEY_SET_VALUE access.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        sub_key: Is a string that names the subkey with which the
            value is associated.
        type: is an integer that specifies the type of the
            data. Currently this must be REG_SZ, meaning only
            strings are supported. Use the ``SetValueEx()`` function
            for support for other data types.
        value: Is a string that specifies the new value.
    Exception:
        ``TypeError``, ``WindowsError`` or ``FileNotFileError``
    """

    test_string(sub_key)
    test_string(value)
    if type != REG_SZ:
        raise TypeError("Type must be wslwinreg.REG_SZ")

    buffer = struct.pack(
        '<BQ',
        Commands.SET_VALUE.value,
        PyHKEY.make(key).hkey)

    _CONNECTION_SOCKET.sendall(
        buffer +
        create_string_buffer(sub_key) +
        create_string_buffer(value))

    # Error code
    handleLRESULT()

########################################


def SetValueEx(key, value_name, reserved, type, value):
    """
    Stores data in the value field of an open registry key.

    This method can also set additional value and type information for the
    specified key.  The key identified by the key parameter must have been
    opened with KEY_SET_VALUE access.

    To open the key, use the CreateKeyEx() or OpenKeyEx() methods.

    Value lengths are limited by available memory. Long values (more than
    2048 bytes) should be stored as files with the filenames stored in
    the configuration registry.  This helps the registry perform efficiently.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        value_name: Is a string that names the subkey with which
            the value is associated.
        reserved: can be anything – zero is always passed to the API.
        type: Is an integer that specifies the type of the data.
        value: Is a string that specifies the new value.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    test_string(value_name)
    temp_buf = to_registry_bytes(value, type)
    buffer = struct.pack(
        '<BQI',
        Commands.SET_VALUE_EX.value,
        PyHKEY.make(key).hkey,
        type)

    _CONNECTION_SOCKET.sendall(
        buffer + create_string_buffer(value_name) +
        create_string_buffer(temp_buf.raw, True))

    # Error code
    handleLRESULT()

########################################


def DisableReflectionKey(key):
    """
    Disables registry reflection.

    Disables registry reflection for 32-bit processes running on a
    64-bit operating system.

    If the key is not on the reflection list, the function succeeds
    but has no effect. Disabling reflection for a key does not
    affect reflection of any subkeys.

    Args:
        key: Is an already open key, or one of the predefined
            HKEY_* constants.
    Exception:
        ``WindowsError``
    """

    # Send the command and a QWORD of the pointer
    buffer = struct.pack(
        '<BQ',
        Commands.DISABLE_REFLECTION_KEY.value,
        PyHKEY.make(key).hkey)
    _CONNECTION_SOCKET.sendall(buffer)

    handleLRESULT()

########################################


def EnableReflectionKey(key):
    """
    Restores registry reflection for the specified disabled key.

    Restoring reflection for a key does not affect reflection of any subkeys.

    Args:
        key: Is an already open key, or one of the predefined
            HKEY_* constants.
    Exception:
        ``WindowsError``
    """

    # Send the command and a QWORD of the pointer
    buffer = struct.pack(
        '<BQ',
        Commands.ENABLE_REFLECTION_KEY.value,
        PyHKEY.make(key).hkey)
    _CONNECTION_SOCKET.sendall(buffer)

    handleLRESULT()

########################################


def QueryReflectionKey(key):
    """
    Determines the reflection state for the specified key.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
    Returns:
        ``True`` if reflection is disabled.
    """

    # Send the command and a QWORD of the pointer
    buffer = struct.pack(
        '<BQ',
        Commands.QUERY_REFLECTION_KEY.value,
        PyHKEY.make(key).hkey)
    _CONNECTION_SOCKET.sendall(buffer)

    # Get the LRESULT
    data = _CONNECTION_SOCKET.recv(1)
    handleLRESULT()
    # Error code
    return_code = struct.unpack('<B', data)[0]
    return return_code != 0

########################################


def convert_to_windows_path(path_name):
    """
    Convert a WSL path to windows if needed.

    If the path is already Windows format, it will be returned unchanged.

    Args:
        path_name: Windows or Linux pathname
    Return:
        Pathname converted to Windows.
    See Also:
        convert_from_windows_path
    """

    # Network drive name?
    if path_name.startswith('\\\\') or ':' in path_name:
        return path_name

    # The tool doesn't process ~ properly, help it by preprocessing here.
    args = ('wslpath',
        '-a',
        '-w',
        os.path.abspath(os.path.expanduser(path_name))
            )

    # Perform the conversion
    tempfp = subprocess.Popen(args, stdout=subprocess.PIPE,
                              stderr=None, universal_newlines=True)
    # Get the string returned by cygpath
    stdoutstr, _ = tempfp.communicate()

    # Error? Fail
    if tempfp.returncode:
        return None
    return stdoutstr.strip()

########################################


def convert_from_windows_path(path_name):
    """
    Convert an absolute Windows path to WSL.

    If the path is already Linux format, it will be returned unchanged.

    Args:
        path_name: Absolute Windows pathname
    Return:
        Pathname converted to Linux.
    See Also:
        convert_to_windows_path
    """

    # Network drive name?
    if path_name[0] in ('~', '/'):
        return path_name

    # Create command list
    args = ('wslpath', '-a', '-u', path_name)

    # Perform the conversion
    tempfp = subprocess.Popen(args, stdout=subprocess.PIPE,
                              stderr=None, universal_newlines=True)
    # Get the string returned by cygpath
    stdoutstr, _ = tempfp.communicate()

    # Error? Fail
    if tempfp.returncode:
        return None
    return stdoutstr.strip()

########################################


def get_file_info(path_name, string_name):
    r"""
    Extract information from a windows exe file version resource.

    Given a windows exe file, extract the 'StringFileInfo' resource and
    parse out the data chunk named by string_name.

    Full list of resource names:
        https://docs.microsoft.com/en-us/windows/desktop/menurc/stringfileinfo-block

    Examples:
        file_version = get_file_info('devenv.exe', 'FileVersion')
        product_version =  get_file_info('devenv.exe', 'ProductVersion')

    Args:
        path_name: Name of the windows file.
        string_name: Name of the data chunk to retrieve

    Return:
        None if no record found or an error, or a valid string
    """

    # Sanity check
    test_string(path_name)
    test_string(string_name)

    # Handle import for Cygwin
    path_name = convert_to_windows_path(path_name)

    # Send the command and the strings
    buffer = struct.pack(
        '<B',
        Commands.GET_FILE_INFO.value)

    _CONNECTION_SOCKET.sendall(
        buffer + create_string_buffer(path_name) +
        create_string_buffer(string_name))

    new_data = recv_string()
    handleLRESULT()
    return new_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains stub functions for unsupported platforms.
"""

## \package wslwinreg.nullapi

# Disable camel case requirement for function names
# pylint: disable=invalid-name

# Disable reusing reserved words.
# pylint: disable=redefined-builtin
# pylint: disable=unused-argument

from .common import KEY_WRITE, KEY_WOW64_64KEY, KEY_READ

## Shared ``NotImplementedError`` for this module
_NOT_IMPL = NotImplementedError(
    'Only supported under Cygwin, MSYS2, or Windows Subsystem for Linux')

########################################


def CloseKey(hkey):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def ConnectRegistry(computer_name, key):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def CreateKey(key, sub_key):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def CreateKeyEx(key, sub_key, reserved=0, access=KEY_WRITE):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL


def DeleteKey(key, sub_key):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def DeleteKeyEx(key, sub_key, access=KEY_WOW64_64KEY, reserved=0):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def DeleteValue(key, value):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def EnumKey(key, index):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def EnumValue(key, index):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def ExpandEnvironmentStrings(str):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def FlushKey(key):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def LoadKey(key, sub_key, file_name):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def OpenKey(key, sub_key, reserved=0, access=KEY_READ):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def OpenKeyEx(key, sub_key, reserved=0, access=KEY_READ):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def QueryInfoKey(key):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def QueryValue(key, sub_key):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def QueryValueEx(key, value_name):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def SaveKey(key, file_name):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def SetValue(key, sub_key, type, value):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def SetValueEx(key, value_name, reserved, type, value):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def DisableReflectionKey(key):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def EnableReflectionKey(key):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def QueryReflectionKey(key):
    """
    Not implemented.

    Exception:
        ``NotImplementedError`` is always thrown.
    """

    raise _NOT_IMPL

########################################


def convert_to_windows_path(path_name):
    """
    Convert pathname to Windows.

    This is the null function, it returns the path_name unchanged.

    Args:
        path_name: Absolute Windows pathname
    Return:
        path_name as is.
    """
    return path_name


########################################


def convert_from_windows_path(path_name):
    """
    Convert pathname from Windows.

    This is the null function, it returns the path_name unchanged.

    Args:
        path_name: Absolute Windows pathname
    Return:
        path_name as is.
    """
    return path_name

########################################


def get_file_info(path_name, string_name):
    """
    Not implemented.

    Args:
        path_name: Name of the windows file.
        string_name: Name of the data chunk to retrieve

    Return:
        None
    """

    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that implements winreg for MSYS and Cygwin
"""

## \package wslwinreg.cygwinapi

# Disable camel case requirement for function names
# pylint: disable=invalid-name
# pylint: disable=useless-object-inheritance
# pylint: disable=unused-argument
# pylint: disable=broad-except

from re import sub as re_sub
import array
import os.path
import subprocess
from ctypes import cdll, create_unicode_buffer, c_void_p, c_ulong, byref, \
    cast, sizeof, create_string_buffer, wstring_at, string_at

from .common import PY2, builtins, ERROR_SUCCESS, ERROR_FILE_NOT_FOUND, \
    ERROR_MORE_DATA, KEY_WOW64_64KEY, KEY_WRITE, KEY_READ, REG_SZ, \
    FORMAT_MESSAGE_ALLOCATE_BUFFER, FORMAT_MESSAGE_IGNORE_INSERTS, \
    FORMAT_MESSAGE_FROM_SYSTEM, LANG_NEUTRAL, LPCVOID, LPVOID, DWORD, PDWORD, \
    LPDWORD, LONG, PLONG, PBYTE, LPBYTE, LPWSTR, LPCWSTR, HKEY, PHKEY, \
    HLOCAL, REGSAM, FILETIME, PFILETIME, SUBLANG_DEFAULT, \
    to_registry_bytes, from_registry_bytes, winerror_to_errno, BOOL

# Test kernel32 in case cdll is the broken version
try:
    cdll.kernel32
except OSError:
    from ctypes import CDLL

    ## Loaded instance of the Windows dll kernel32
    cdll.kernel32 = CDLL("Kernel32.dll", use_errno=True)

    ## Loaded instance of the Windows dll advapi32
    cdll.advapi32 = CDLL("advapi32.dll", use_errno=True)

    ## Loaded instance of the Windows dll version
    cdll.version = CDLL('version.dll', use_errno=True)

## Type long for Python 2 compatibility
try:
    long
except NameError:
    # Fake it for Python 3
    long = int

########################################

# Windows functions extracted from cdll

## WINBASEAPI DWORD WINAPI FormatMessageW(DWORD,LPCVOID,DWORD,DWORD,LPWSTR,
#                                         DWORD, va_list*)
FormatMessageW = cdll.kernel32.FormatMessageW
FormatMessageW.restype = DWORD
FormatMessageW.argtypes = [DWORD, LPCVOID, DWORD, DWORD, LPCVOID, DWORD,
                           c_void_p]

## WINBASEAPI DWORD WINAPI GetLastError(void);
GetLastError = cdll.kernel32.GetLastError
GetLastError.restype = DWORD
GetLastError.argtypes = []

## WINBASEAPI HLOCAL WINAPI LocalFree(HLOCAL);
LocalFree = cdll.kernel32.LocalFree
LocalFree.restype = HLOCAL
LocalFree.argtypes = [HLOCAL]

## WINBASEAPI DWORD WINAPI ExpandEnvironmentStringsW(LPCWSTR lpSrc, LPWSTR
#                                                   lpDst, DWORD nSize);
ExpandEnvironmentStringsW = cdll.kernel32.ExpandEnvironmentStringsW
ExpandEnvironmentStringsW.restype = DWORD
ExpandEnvironmentStringsW.argtypes = [LPCWSTR, LPWSTR, DWORD]

## WINADVAPI LONG WINAPI RegCloseKey(HKEY);
RegCloseKey = cdll.advapi32.RegCloseKey
RegCloseKey.restype = LONG
RegCloseKey.argtypes = [HKEY]

## WINADVAPI LONG WINAPI RegConnectRegistryW(LPCWSTR,HKEY,PHKEY);
RegConnectRegistryW = cdll.advapi32.RegConnectRegistryW
RegConnectRegistryW.restype = LONG
RegConnectRegistryW.argtypes = [LPCWSTR, HKEY, PHKEY]

## WINADVAPI LONG WINAPI RegCreateKeyW(HKEY,LPCWSTR,PHKEY);
RegCreateKeyW = cdll.advapi32.RegCreateKeyW
RegCreateKeyW.restype = LONG
RegCreateKeyW.argtypes = [HKEY, LPCWSTR, PHKEY]

## WINADVAPI LONG WINAPI RegCreateKeyW(HKEY,LPCWSTR,DWORD, LPWSTR, DWORD,
#                                     REGSAM, LPCVOID, PHKEY, LPDWORD);
RegCreateKeyExW = cdll.advapi32.RegCreateKeyExW
RegCreateKeyExW.restype = LONG
RegCreateKeyExW.argtypes = [
    HKEY,
    LPCWSTR,
    DWORD,
    LPWSTR,
    DWORD,
    REGSAM,
    LPCVOID,
    PHKEY,
    LPDWORD]

## WINADVAPI LONG WINAPI RegDeleteKeyW(HKEY,LPCWSTR);
RegDeleteKeyW = cdll.advapi32.RegDeleteKeyW
RegDeleteKeyW.restype = LONG
RegDeleteKeyW.argtypes = [HKEY, LPCWSTR]

## WINADVAPI LONG WINAPI RegDeleteKeyW(HKEY,LPCWSTR,REGSAM,DWORD);
RegDeleteKeyExW = cdll.advapi32.RegDeleteKeyExW
RegDeleteKeyExW.restype = LONG
RegDeleteKeyExW.argtypes = [HKEY, LPCWSTR, REGSAM, DWORD]

## WINADVAPI LONG WINAPI RegDeleteValueW(HKEY,LPCWSTR);
RegDeleteValueW = cdll.advapi32.RegDeleteValueW
RegDeleteValueW.restype = LONG
RegDeleteValueW.argtypes = [HKEY, LPCWSTR]

## WINADVAPI LONG WINAPI RegEnumKeyExW(HKEY,DWORD,LPWSTR,PDWORD,PDWORD,
#                                     LPWSTR,PDWORD,PFILETIME);
RegEnumKeyExW = cdll.advapi32.RegEnumKeyExW
RegEnumKeyExW.restype = LONG
RegEnumKeyExW.argtypes = [HKEY, DWORD, LPWSTR, PDWORD, PDWORD, LPWSTR,
                          PDWORD, c_void_p]

## WINADVAPI LONG WINAPI RegQueryInfoKeyW(HKEY,LPWSTR,PDWORD,PDWORD,PDWORD,
#                                        PDWORD,PDWORD,PDWORD,PDWORD,PDWORD,
#                                        PDWORD,PFILETIME);
RegQueryInfoKeyW = cdll.advapi32.RegQueryInfoKeyW
RegQueryInfoKeyW.restype = LONG
RegQueryInfoKeyW.argtypes = [HKEY, LPWSTR, PDWORD, PDWORD, PDWORD, PDWORD,
                             PDWORD, PDWORD, PDWORD, PDWORD, PDWORD, PFILETIME]

## WINADVAPI LONG WINAPI RegEnumValueW(HKEY,DWORD,LPWSTR,PDWORD,PDWORD,PDWORD,
#                                     LPBYTE,PDWORD);
RegEnumValueW = cdll.advapi32.RegEnumValueW
RegEnumValueW.restype = LONG
RegEnumValueW.argtypes = [HKEY, DWORD, LPWSTR, PDWORD, PDWORD, PDWORD,
                          LPBYTE, PDWORD]

## WINADVAPI LONG WINAPI RegFlushKey(HKEY);
RegFlushKey = cdll.advapi32.RegFlushKey
RegFlushKey.restype = LONG
RegFlushKey.argtypes = [HKEY]

## WINADVAPI LONG WINAPI RegLoadKeyW(HKEY,LPCWSTR,LPCWSTR)
RegLoadKeyW = cdll.advapi32.RegLoadKeyW
RegLoadKeyW.restype = LONG
RegLoadKeyW.argtypes = [HKEY, LPCWSTR, LPCWSTR]

## WINADVAPI LONG WINAPI RegOpenKeyExW(HKEY,LPCWSTR,DWORD,REGSAM,PHKEY)
RegOpenKeyExW = cdll.advapi32.RegOpenKeyExW
RegOpenKeyExW.restype = LONG
RegOpenKeyExW.argtypes = [HKEY, LPCWSTR, DWORD, REGSAM, PHKEY]

## WINADVAPI LONG RegQueryReflectionKey(HKEY, LPDWORD)
RegQueryReflectionKey = cdll.advapi32.RegQueryReflectionKey
RegQueryReflectionKey.restype = LONG
RegQueryReflectionKey.argtypes = [HKEY, PDWORD]

## WINADVAPI LONG RegDisbleReflectionKey(HKEY)
RegDisableReflectionKey = cdll.advapi32.RegDisableReflectionKey
RegDisableReflectionKey.restype = LONG
RegDisableReflectionKey.argtypes = [HKEY]

## WINADVAPI LONG RegEnableReflectionKey(HKEY)
RegEnableReflectionKey = cdll.advapi32.RegEnableReflectionKey
RegEnableReflectionKey.restype = LONG
RegEnableReflectionKey.argtypes = [HKEY]

## WINADVAPI LONG WINAPI RegQueryValueW(HKEY,LPCWSTR,LPWSTR,PLONG);
RegQueryValueW = cdll.advapi32.RegQueryValueW
RegQueryValueW.restype = LONG
RegQueryValueW.argtypes = [HKEY, LPCWSTR, LPWSTR, PLONG]

## WINADVAPI LONG WINAPI RegQueryValueExW(HKEY,LPCWSTR,LPDWORD,LPDWORD,LPBYTE,
#                                        LPDWORD);
RegQueryValueExW = cdll.advapi32.RegQueryValueExW
RegQueryValueExW.restype = LONG
RegQueryValueExW.argtypes = [HKEY, LPCWSTR, LPDWORD, LPDWORD, LPBYTE, LPDWORD]

## WINADVAPI LONG WINAPI RegSaveKeyW(HKEY,LPCWSTR,LPSECURITY_ATTRIBUTES);
RegSaveKeyW = cdll.advapi32.RegSaveKeyW
RegSaveKeyW.restype = LONG
RegSaveKeyW.argtypes = [HKEY, LPCWSTR, c_void_p]

## WINADVAPI LONG WINAPI RegSetValueW(HKEY,LPCWSTR,DWORD,LPCWSTR,DWORD);
RegSetValueW = cdll.advapi32.RegSetValueW
RegSetValueW.restype = LONG
RegSetValueW.argtypes = [HKEY, LPCWSTR, DWORD, LPCWSTR, DWORD]

## WINADVAPI LONG WINAPI RegSetValueExW(HKEY,LPCWSTR,DWORD,DWORD,const BYTE*,
#                                      DWORD);
RegSetValueExW = cdll.advapi32.RegSetValueExW
RegSetValueExW.restype = LONG
RegSetValueExW.argtypes = [HKEY, LPCWSTR, DWORD, DWORD, PBYTE, DWORD]

## WINADVAPI DWORD WINAPI GetFileVersionInfoSizeW(LPCWSTR, LPDWORD);
GetFileVersionInfoSizeW = cdll.version.GetFileVersionInfoSizeW
GetFileVersionInfoSizeW.restype = DWORD
GetFileVersionInfoSizeW.argtypes = [LPCWSTR, LPDWORD]

## WINADVAPI BOOL WINAPI GetFileVersionInfoW(LPCWSTR, DWORD, DWORD, LPVOID);
GetFileVersionInfoW = cdll.version.GetFileVersionInfoW
GetFileVersionInfoW.restype = BOOL
GetFileVersionInfoW.argtypes = [LPCWSTR, DWORD, DWORD, LPVOID]

## WINADVAPI BOOL WINAPI VerQueryValueW(LPCVOID, LPCWSTR, LPVOID, PUINT);
VerQueryValueW = cdll.version.VerQueryValueW
VerQueryValueW.restype = BOOL
VerQueryValueW.argtypes = [LPCVOID, LPCWSTR, LPVOID, PLONG]

########################################


def winerror_to_string(winerror):
    """
    Convert a windows integer error code into a descriptive string.

    Call FormatMessageW() in Windows to perform the actual work.

    Args:
        winerror: Integer error code from Windows

    Returns:
        String describing the error.
    """

    # buf is actually a string, but is initially set to NULL
    buf = c_void_p(None)
    msg_len = FormatMessageW(
        # Error API error
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        # Message source
        None,
        c_ulong(winerror),
        # Use the user's default language
        (LANG_NEUTRAL<<10) + SUBLANG_DEFAULT,
        byref(buf),
        # Size is not used
        0,
        # No arguments
        None)

    if msg_len == 0:
        # Only seen this in out of mem situations
        return "Windows Error 0x%X" % winerror

    result = cast(buf, LPWSTR).value
    LocalFree(buf)

    # Clean up the string before returning
    return re_sub(r'[\s.]*$', '', result)

########################################


def check_LRESULT(return_code):
    """
    Test LRESULT for an error code and raise exception on error condition.

    If the input value is zero, return the value. If the input is a non-zero
    code, raise a WindowsError() exception unless it's ERROR_FILE_NOT_FOUND.
    In the latter case, raise a ``FileNotFoundError``.

    Args:
        return_code: Windows integer error code.
    Returns:
        Error code, usually zero.
    Exception:
        ``WindowsError`` or ``FileNotFoundError``
    """

    # Make sure it's not a ctype class
    if not isinstance(return_code, (int, long)):
        return_code = return_code.value

    # Was there an error?
    if return_code != ERROR_SUCCESS:
        # Special case, unit tests require a FileNotFoundError
        if return_code == ERROR_FILE_NOT_FOUND:
            raise FileNotFoundError(winerror_to_string(return_code))

        # Throw a WindowsError exception
        raise WindowsError(return_code)

    # Return the zero.
    return return_code

########################################


# pylint: disable=redefined-builtin

class WindowsError(OSError):
    """
    This exception doesn't exist in Cygwin/MSYS, provide it.
    """

    # pylint: disable=super-init-not-called
    def __init__(self, winerror, strerror=None, filename=None):
        """
        Initialize a WindowsError exception

        Args:
            winerror: The windows error code
            strerror: The string describing the error
            filename: Name of the file that caused this error, if applicable.
        """

        # Check if an error was provided
        if winerror == 0:
            # Call GetLastError() in windows to make sure
            winerror = GetLastError()

        ## Windows error code
        self.winerror = winerror

        ## Linux style error code
        self.errno = winerror_to_errno(winerror)

        if strerror is None:
            strerror = winerror_to_string(winerror)

        ## Description string of the error
        self.strerror = strerror

        ## Name of the responsible file for this error
        self.filename = filename

    def __str__(self):
        """
        Convert the class into a string.

        Returns:
            String describing the error contained.
        """
        if self.filename is not None:
            return "[Error %s] %s: %s" % (self.winerror, self.strerror,
                                          repr(self.filename))
        return "[Error %s] %s" % (self.winerror, self.strerror)


# Ensure this exception is available globally.
builtins.WindowsError = WindowsError

########################################


class PyHKEY(object):
    """
    A Python object representing a win32 registry key.

    This object wraps a Windows HKEY object, automatically closing it when the
    object is destroyed. To guarantee cleanup, you can call either the Close()
    method on the object, or the CloseKey() function.

    All registry functions in this module return one of these objects.

    All registry functions in this module which accept a handle object also
    accept an integer, however, use of the handle object is encouraged.

    Handle objects provide semantics for __bool__() – thus

    @code
        if handle:
            print("Yes")
    @endcode

    will print Yes if the handle is currently valid (has not been closed or
    detached).

    The object also support comparison semantics, so handle objects will compare
    true if they both reference the same underlying Windows handle value.

    Handle objects can be converted to an integer (e.g., using the built-in
    int() function), in which case the underlying Windows handle value is
    returned. You can also use the Detach() method to return the integer handle,
    and also disconnect the Windows handle from the handle object.
    """

    def __init__(self, hkey, null_ok=False):
        """
        Initialize the PyHKEY class.

        Args:
            hkey: Integer value representing the pointer to the HKEY
            null_ok: True if None is acceptable.
        """
        if not isinstance(hkey, (int, long)):
            if not null_ok or not isinstance(hkey, type(None)):
                raise TypeError("A handle must be an integer")

        ## Integer that represents the HANDLE pointer
        self.hkey = hkey

    def __del__(self):
        """
        Called when this object is garbage collected.
        """
        if self.hkey:
            # Ignore errors.
            try:
                self.Close()
            except BaseException:
                pass

    def Close(self):
        """
        Closes the underlying Windows handle.

        Note:
            If the handle is already closed, no error is raised.
        """
        if self.hkey:
            check_LRESULT(RegCloseKey(self.hkey))
        self.hkey = 0

    def Detach(self):
        """
        Detaches the Windows handle from the handle object.

        Note:
        The result is the value of the handle before it is detached. If the
        handle is already detached, this will return zero.

        After calling this function, the handle is effectively invalidated,
        but the handle is not closed.  You would call this function when you
        need the underlying win32 handle to exist beyond the lifetime of the
        handle object.

        On 64 bit windows, the result of this function is a long integer.

        Returns:
            Previous HKEY integer.
        """
        hkey = self.hkey
        self.hkey = 0
        return hkey

    def _handle(self):
        """
        The integer Win32 handle.
        """
        return self.hkey

    # The integer Win32 handle.
    handle = property(_handle)

    def __enter__(self):
        """
        Called when object is entered.

        Note:
            Needed for the Python ``with`` statement.
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Release handle on class destruction.

        Note:
            Needed for the Python ``with`` statement.
        """
        self.Close()
        return False

    def __hash__(self):
        """
        Convert the object into a hash.

        Note:
            The implementation returns the id, which should be
            random enough.
        """
        return id(self)

    def __int__(self):
        """
        Converting a handle to an integer returns the Win32 handle.
        """
        return self.hkey

    if PY2:
        def __nonzero__(self):
            """
            Handles with an open object return true, otherwise false.
            """
            return bool(self.hkey)
    else:
        def __bool__(self):
            """
            Handles with an open object return true, otherwise false.
            """
            return bool(self.hkey)

    def __repr__(self):
        """
        Return descriptive string for the class object.
        """
        return "<PyHKEY at %08X (%08X)>" % (id(self), self.hkey)

    def __str__(self):
        """
        Return short string for the handle object.
        """
        return "<PyHKEY:%08X>" % self.hkey

    # Python sort description.
    _as_parameter_ = property(lambda self: self.hkey)

    @staticmethod
    def make(hkey, null_ok=None):
        """
        Convert a pointer into a PyHKEY object.
        """

        # Already converted?
        if isinstance(hkey, PyHKEY):
            return hkey

        # Is it a valid integer?
        if isinstance(hkey, (int, long)):
            return PyHKEY(hkey)

        # None, if None is allowed.
        if null_ok and isinstance(hkey, type(None)):
            return PyHKEY(None)

        # This didn't end well.
        raise TypeError("A handle must be a HKEY object or an integer")


########################################


def CloseKey(hkey):
    """
    Closes a previously opened registry key.

    The hkey argument specifies a previously opened key.

    Args:
        hkey: PyHKEY object or None.

    Note:
        If hkey is not closed using this method (or via hkey.Close()),
        it is closed when the hkey object is destroyed by Python.

    """
    PyHKEY.make(hkey).Close()

########################################


def ConnectRegistry(computer_name, key):
    """
    Establishes a connection to a predefined registry handle.

    Establishes a connection to a predefined registry handle on another
    computer, and returns a handle object.

    Args:
        computer_name: Is the name of the remote computer, of the form
            ``r"\\computername"``. If ``None``, the local computer is used.
        key: Is the predefined handle to connect to.
    Returns:
        PyHKEY object
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    result = HKEY()
    rc = RegConnectRegistryW(computer_name, PyHKEY.make(key), byref(result))
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)
    return PyHKEY(result.value)

########################################


def CreateKey(key, sub_key):
    """
    Creates or opens the specified key.

    If key is one of the predefined keys, sub_key may be None. In that case,
    the handle returned is the same key handle passed in to the function.

    If the key already exists, this function opens the existing key.

    Args:
        key: Is an already open key, or one of the predefined HKEY_* constants.
        sub_key: Is a string that names the key this method opens or creates.
    Returns:
        Handle of the opened key.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    result = HKEY()
    rc = RegCreateKeyW(PyHKEY.make(key), sub_key, byref(result))
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)
    return PyHKEY(result.value)

########################################


def CreateKeyEx(key, sub_key, reserved=0, access=KEY_WRITE):
    """
    Creates or opens the specified key.

    If key is one of the predefined keys, sub_key may be None. In that case,
    the handle returned is the same key handle passed in to the function.

    If the key already exists, this function opens the existing key.

    Args:
        key: Is an already open key, or one of the predefined HKEY_* constants
        sub_key: Is a string that names the key this method opens or creates.
        reserved: Is a reserved integer, and must be zero. The default is zero.
        access: Os an integer that specifies an access mask that describes the
            desired security access for the key. Default is common.KEY_WRITE.

    Returns:
        Handle of the opened key.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    result = HKEY()
    rc = RegCreateKeyExW(
            PyHKEY.make(key),
            sub_key,
            reserved,
            None,
            0,
            access,
            None,
            byref(result),
            None)
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)
    return PyHKEY(result.value)

########################################


def DeleteKey(key, sub_key):
    """
    Deletes the specified key.

    If the method succeeds, the entire key, including all of its values,
    is removed.

    Note:
        This method can not delete keys with subkeys.

    Args:
        key: Is an already open key, or one of the predefined HKEY_* constants.
        sub_key: Is a string that must be a subkey of the key identified by the
            key parameter. This value must not be ``None``, and the key may not
            have subkeys.
    Exception:
        ``WindowsError``
    """

    rc = RegDeleteKeyW(PyHKEY.make(key), sub_key)
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)

########################################


def DeleteKeyEx(key, sub_key, access=KEY_WOW64_64KEY, reserved=0):
    """
    Deletes the specified key.

    If the method succeeds, the entire key, including all of its values,
    is removed.

    Note:
        This method can not delete keys with subkeys.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.

        sub_key: Os a string that must be a subkey of the key identified by the
            key parameter. This value must not be ``None``, and the key may not
            have subkeys.

        access: Is an integer that specifies an access mask that describes the
            desired security access for the key. Default
            is common.KEY_WOW64_64KEY.

        reserved: Is a reserved integer, and must be zero. The default is zero.
    Exception:
        ``WindowsError``
    """

    rc = RegDeleteKeyExW(PyHKEY.make(key), sub_key, access, reserved)
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)

########################################


def DeleteValue(key, value):
    """
    Removes a named value from a registry key.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        value: Is a string that identifies the value to remove.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    rc = RegDeleteValueW(PyHKEY.make(key), value)
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)

########################################


def EnumKey(key, index):
    """
    Enumerates subkeys of an open registry key, returning a string.

    The function retrieves the name of one subkey each time it is called.
    It is typically called repeatedly until an ``OSError`` exception is
    raised, indicating no more values are available.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        index: Is an integer that identifies the index of the key to retrieve.

    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    # max key name length is 256 if unterminated
    tmpbuf = create_unicode_buffer(257)
    length = DWORD(sizeof(tmpbuf))
    rc = RegEnumKeyExW(PyHKEY.make(key), index, tmpbuf, byref(length),
                          None, None, None, None)
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)
    return tmpbuf[:length.value]

########################################


def EnumValue(key, index):
    # pylint: disable=line-too-long
    """
    Enumerates values of an open registry key, returning a tuple.

    The function retrieves the name of one subkey each time it is called.
    It is typically called repeatedly, until an ``OSError`` exception is
    raised, indicating no more values.

    | Index | Meaning |
    | ----- | ------- |
    | 0 | A string that identifies the value. |
    | 1 | An object that holds the value data, and whose type depends on the underlying registry type |
    | 2 | An integer that identifies the type of the value data |

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        index: Is an integer that identifies the index of the value to retrieve.
    Returns:
       A tuple of 3 items.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    # Check the key
    hkey = PyHKEY.make(key)

    # Get information from the key
    value_size = DWORD()
    data_size = DWORD()
    check_LRESULT(
        RegQueryInfoKeyW(
            hkey,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            byref(value_size),
            byref(data_size),
            None,
            None))

    # Include null terminators
    value_size.value += 1
    data_size.value += 1

    value_buf = create_unicode_buffer(value_size.value)
    data_buf = create_string_buffer(data_size.value)
    save_data_size = data_size.value
    save_value_size = value_size.value
    typ = DWORD()
    while True:
        rc = RegEnumValueW(hkey, index, value_buf, byref(value_size), None,
                          byref(typ), data_buf, byref(data_size))
        if rc != ERROR_MORE_DATA:
            break
        del data_buf
        save_data_size = save_data_size * 2
        data_size.value = save_data_size
        data_buf = create_string_buffer(save_data_size)
        value_size.value = save_value_size

    # Throw if the error code is not ERROR_SUCCESS
    check_LRESULT(rc)
    return (value_buf[:value_size.value],
            from_registry_bytes(data_buf, data_size, typ),
            typ.value)

########################################


def ExpandEnvironmentStrings(str):
    """
    Expands environment variables.

    Expands environment variable placeholders %NAME% in strings
    like REG_EXPAND_SZ.

    Args:
        str: String to expand.
    Returns:
        Expanded string
    Exception:
        ``WindowsError``
    """

    # Determine the size of the return value
    result_size = ExpandEnvironmentStringsW(str, None, 0)
    # If the return value is zero, there was an error
    if not result_size:
        raise WindowsError(0)

    # Make a buffer and call the function again
    result_size += 1
    result_buffer = create_unicode_buffer(result_size)
    length = DWORD(sizeof(result_buffer))
    result_size = ExpandEnvironmentStringsW(str, result_buffer, length)
    if not result_size:
        raise WindowsError(0)
    return result_buffer.value

########################################


def FlushKey(key):
    """
    Writes all the attributes of a key to the registry.

    It is not necessary to call ``FlushKey()`` to change a key. Registry
    changes are flushed to disk by the registry using its lazy flusher.
    Registry changes are also flushed to disk at system shutdown. Unlike
    ``CloseKey()``, the ``FlushKey()`` method returns only when all the
    data has been written to the registry. An application should only
    call ``FlushKey()`` if it requires absolute certainty that registry
    changes are on disk.

    Note:
        If you don’t know whether a ``FlushKey()`` call is required, it
        probably isn’t.
    Args:
        key: Is an already open key, or any one of the predefined
        HKEY_* constants.
    Exception:
        ``WindowsError``
    """

    rc = RegFlushKey(PyHKEY.make(key))
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)

########################################


def LoadKey(key, sub_key, file_name):
    """
    Creates a subkey under the specified key.

    Creates a subkey under the specified key and stores registration
    information from a specified file into that subkey.

    This file must have been created with the SaveKey() function.
    Under the file allocation table (FAT) file system, the filename may not
    have an extension.

    A call to ``LoadKey()`` fails if the calling process does not have the
    SE_RESTORE_PRIVILEGE privilege.

    If key is a handle returned by ``ConnectRegistry()``, then the path
    specified in fileName is relative to the remote computer.

    Args:
        key: Is a handle returned by ``ConnectRegistry()`` or one of the
            constants common.HKEY_USERS or common.HKEY_LOCAL_MACHINE.
        sub_key: Is a string that identifies the sub_key to load
        file_name: Is the name of the file to load registry data from.

    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    rc = RegLoadKeyW(PyHKEY.make(key), sub_key, file_name)
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)

########################################


def OpenKey(key, sub_key, reserved=0, access=KEY_READ):
    """
    Opens the specified key, returning a handle object.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        sub_key: Is a string that identifies the sub_key to open
        reserved: Is a reserved integer, and must be zero.
            Default is zero.
        access: Is an integer that specifies an access mask that
            describes the desired security access for the key. Default
            is KEY_READ.

    Returns:
        A new handle to the specified key.

    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    result = HKEY()
    rc = RegOpenKeyExW(
            PyHKEY.make(key),
            sub_key,
            reserved,
            access,
            byref(result))
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)
    return PyHKEY.make(result.value)

########################################


def OpenKeyEx(key, sub_key, reserved=0, access=KEY_READ):
    """
    Opens the specified key, returning a handle object.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        sub_key: Is a string that identifies the sub_key to open
        reserved: Is a reserved integer, and must be zero.
            Default is zero.
        access: Is an integer that specifies an access mask that
            describes the desired security access for the key. Default
            is KEY_READ.

    Returns:
        A new handle to the specified key.

    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """
    return OpenKey(key, sub_key, reserved, access)

########################################


def QueryInfoKey(key):
    # pylint: disable=line-too-long
    """
    Returns information about a key, as a tuple.

    | Index | Meaning |
    | ----- | ------- |
    | 0 | An integer giving the number of sub keys this key has. |
    | 1 | An integer giving the number of values this key has. |
    | 2 | An integer giving when the key was last modified (if available) as 100’s of nanoseconds since Jan 1, 1601. |

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
    Returns:
        A tuple of 3 items.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    num_sub_keys = DWORD()
    num_values = DWORD()
    last_write_time = FILETIME()
    check_LRESULT(RegQueryInfoKeyW(PyHKEY.make(key), None, None, None,
                             byref(num_sub_keys), None, None,
                             byref(num_values), None, None, None,
                             byref(last_write_time)))
    last_write_time = (last_write_time.high << 32) | last_write_time.low
    return (num_sub_keys.value, num_values.value, last_write_time)

########################################


def QueryValue(key, sub_key):
    """
    Retrieves the unnamed value for a key, as a string.

    Values in the registry have name, type, and data components. This method
    retrieves the data for a key’s first value that has a NULL name. But the
    underlying API call doesn’t return the type, so always use
    ``QueryValueEx()`` if possible.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        sub_key: Is a string that holds the name of the subkey with which the
            value is associated. If this parameter is None or empty, the
            function retrieves the value set by the ``SetValue()`` method for
            the key identified by ``key``.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    hkey = PyHKEY.make(key)
    buf_size = LONG()
    rc = RegQueryValueW(hkey, sub_key, None, byref(buf_size))
    if rc == ERROR_MORE_DATA:
        buf_size.value = 256
    elif rc != ERROR_SUCCESS:
        check_LRESULT(rc)

    buffer_size = buf_size.value
    buf = create_unicode_buffer(buf_size.value)
    while True:
        rc = RegQueryValueW(hkey, sub_key, buf, byref(buf_size))
        if rc != ERROR_MORE_DATA:
            break
        buffer_size = buffer_size * 2
        del buf
        buf_size.value = buffer_size
        buf = create_unicode_buffer(buf_size.value)

    check_LRESULT(rc)
    return buf.value

########################################


def QueryValueEx(key, value_name):
    """
    Retrieves the type and data for a specified value name.

    Retrieves the type and data for a specified value name associated with an
    open registry key.

    | Index | Meaning |
    | ----- | ------- |
    | 0 | The value of the registry item. |
    | 1 | An integer giving the registry type for this value. |

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        value_name: Is a string indicating the value to query.
    Returns:
        A tuple of 2 items.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    hkey = PyHKEY.make(key)
    buf_size = DWORD()
    rc = RegQueryValueExW(hkey, value_name, None, None, None,
                          byref(buf_size))
    # Set to 256 for looping
    if rc == ERROR_MORE_DATA:
        buf_size.value = 256
    elif rc != ERROR_SUCCESS:
        check_LRESULT(rc)

    # buf is a byte array
    buffer_size = buf_size.value
    buf = create_string_buffer(buf_size.value)
    typ = DWORD()
    while True:
        rc = RegQueryValueExW(hkey, value_name, None, byref(typ), buf,
                             byref(buf_size))
        if rc != ERROR_MORE_DATA:
            break
        buffer_size = buffer_size * 2
        del buf
        buf_size.value = buffer_size
        buf = create_string_buffer(buffer_size)

    check_LRESULT(rc)
    return (from_registry_bytes(buf, buf_size, typ),
            typ.value)

########################################


def SaveKey(key, file_name):
    """
    Saves the specified key, and all its subkeys to the specified file.

    If key represents a key on a remote computer, the path described by
    file_name is relative to the remote computer. The caller of this method
    must possess the SeBackupPrivilege security privilege.

    This function passes NULL for security_attributes to the API.

    Note:
        Privileges are different than permissions.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        file_name: Is the name of the file to save registry data to.
            This file cannot already exist. If this filename includes an
            extension, it cannot be used on file allocation table (FAT) file
            systems by the ``LoadKey()`` method.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    rc = RegSaveKeyW(PyHKEY.make(key), file_name, None)
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)

########################################


def SetValue(key, sub_key, type, value):
    """
    Associates a value with a specified key.

    If the key specified by the sub_key parameter does not exist,
    the SetValue function creates it.

    Value lengths are limited by available memory. Long values
    (more than 2048 bytes) should be stored as files with the
    filenames stored in the configuration registry.
    This helps the registry perform efficiently.

    The key identified by the ``key`` parameter must have been
    opened with common.KEY_SET_VALUE access.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        sub_key: Is a string that names the subkey with which the
            value is associated.
        type: is an integer that specifies the type of the
            data. Currently this must be REG_SZ, meaning only
            strings are supported. Use the ``SetValueEx()`` function
            for support for other data types.
        value: Is a string that specifies the new value.
    Exception:
        ``TypeError``, ``WindowsError`` or ``FileNotFileError``
    """

    if type != REG_SZ:
        raise TypeError("Type must be wslwinreg.REG_SZ")
    value_buf = create_unicode_buffer(value)
    rc = RegSetValueW(PyHKEY.make(key), sub_key,
                      REG_SZ, value_buf, sizeof(value_buf))
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)

########################################


def SetValueEx(key, value_name, reserved, type, value):
    """
    Stores data in the value field of an open registry key.

    This method can also set additional value and type information for the
    specified key.  The key identified by the key parameter must have been
    opened with KEY_SET_VALUE access.

    To open the key, use the CreateKeyEx() or OpenKeyEx() methods.

    Value lengths are limited by available memory. Long values (more than
    2048 bytes) should be stored as files with the filenames stored in
    the configuration registry.  This helps the registry perform efficiently.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
        value_name: Is a string that names the subkey with which
            the value is associated.
        reserved: can be anything – zero is always passed to the API.
        type: Is an integer that specifies the type of the data.
        value: Is a string that specifies the new value.
    Exception:
        ``WindowsError`` or ``FileNotFileError``
    """

    temp_buf = to_registry_bytes(value, type)
    rc = RegSetValueExW(PyHKEY.make(key), value_name, 0, type,
                        temp_buf, len(temp_buf))
    del temp_buf
    if rc != ERROR_SUCCESS:
        check_LRESULT(rc)


########################################


def DisableReflectionKey(key):
    """
    Disables registry reflection.

    Disables registry reflection for 32-bit processes running on a
    64-bit operating system.

    If the key is not on the reflection list, the function succeeds
    but has no effect. Disabling reflection for a key does not
    affect reflection of any subkeys.

    Args:
        key: Is an already open key, or one of the predefined
            HKEY_* constants.
    Exception:
        ``WindowsError``
    """

    check_LRESULT(RegDisableReflectionKey(PyHKEY.make(key)))

########################################


def EnableReflectionKey(key):
    """
    Restores registry reflection for the specified disabled key.

    Restoring reflection for a key does not affect reflection of any subkeys.

    Args:
        key: Is an already open key, or one of the predefined
            HKEY_* constants.
    Exception:
        ``WindowsError``
    """

    check_LRESULT(RegEnableReflectionKey(PyHKEY.make(key)))

########################################


def QueryReflectionKey(key):
    """
    Determines the reflection state for the specified key.

    Args:
        key: Is an already open key, or any one of the predefined
            HKEY_* constants.
    Returns:
        ``True`` if reflection is disabled.
    """

    is_reflection_disabled = DWORD()
    check_LRESULT(
        RegQueryReflectionKey(
            PyHKEY.make(key),
            byref(is_reflection_disabled)))
    return is_reflection_disabled != 0

########################################


def convert_to_windows_path(path_name):
    """
    Convert a MSYS/Cygwin path to windows if needed.

    If the path is already Windows format, it will be returned unchanged.

    Args:
        path_name: Windows or Linux pathname
    Return:
        Pathname converted to Windows.
    See Also:
        convert_from_windows_path
    """

    # Network drive name?
    if path_name.startswith('\\\\') or ':' in path_name:
        return path_name

    # The tool doesn't process ~ properly, help it by preprocessing here.
    args = ('cygpath',
        '-a',
        '-w',
        os.path.abspath(os.path.expanduser(path_name))
            )

    # Perform the conversion
    tempfp = subprocess.Popen(args, stdout=subprocess.PIPE,
                              stderr=None, universal_newlines=True)
    # Get the string returned by cygpath
    stdoutstr, _ = tempfp.communicate()

    # Error? Fail
    if tempfp.returncode:
        return None
    return stdoutstr.strip()

########################################


def convert_from_windows_path(path_name):
    """
    Convert an absolute Windows path to Cygwin/MSYS2.

    If the path is already Cygwin/MSYS2 format, it will be returned unchanged.

    Args:
        path_name: Absolute Windows pathname
    Return:
        Pathname converted to Linux.
    See Also:
        convert_to_windows_path
    """

    # Network drive name?
    if path_name[0] in ('~', '/'):
        return path_name

    # Create command list
    args = ('cygpath', '-a', '-u', path_name)

    # Perform the conversion
    tempfp = subprocess.Popen(args, stdout=subprocess.PIPE,
                              stderr=None, universal_newlines=True)
    # Get the string returned by cygpath
    stdoutstr, _ = tempfp.communicate()

    # Error? Fail
    if tempfp.returncode:
        return None
    return stdoutstr.strip()

########################################


def get_file_info(path_name, string_name):
    r"""
    Extract information from a windows exe file version resource.

    Given a windows exe file, extract the 'StringFileInfo' resource and
    parse out the data chunk named by string_name.

    Full list of resource names:
        https://docs.microsoft.com/en-us/windows/desktop/menurc/stringfileinfo-block

    Examples:
        file_version = get_file_info('devenv.exe', 'FileVersion')
        product_version =  get_file_info('devenv.exe', 'ProductVersion')

    Args:
        path_name: Name of the windows file.
        string_name: Name of the data chunk to retrieve

    Return:
        None if no record found or an error, or a valid string
    """

    # Handle import for Cygwin
    path_name = convert_to_windows_path(path_name)

    # Ensure it's unicode
    wchar_filename = LPWSTR(path_name)

    # Call windows to get the data size
    size = GetFileVersionInfoSizeW(wchar_filename, None)

    # Was there no data to return?
    if size:

        # Create buffer for resource data
        res_data = create_string_buffer(size)

        # Extract the file data
        GetFileVersionInfoW(
            wchar_filename, 0, size, res_data)

        # Find the default codepage (Not everything is in English)
        record = LPVOID()
        length = LONG()
        VerQueryValueW(
            res_data,
            '\\VarFileInfo\\Translation',
            byref(record),
            byref(length))
        # Was a codepage found?
        if length.value:

            # Parse out the first found codepage (It's the default
            # language) it's in the form of two 16 bit shorts
            codepages = array.array(
                'H', string_at(
                    record.value, length.value))

            # Extract information from the version using unicode and
            # the proper codepage
            if VerQueryValueW(
                    res_data,
                    '\\StringFileInfo\\{0:04x}{1:04x}\\{2}'.format(
                        codepages[0],
                        codepages[1],
                        string_name),
                    byref(record),
                    byref(length)):
                # Return the final result removing the terminating zero
                return wstring_at(record.value, length.value - 1)
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Build the egg file for wslwinreg for python

setup.py clean
setup.py sdist bdist_wheel
twine upload --verbose dist/*

Copyright 2020-2021 by Rebecca Ann Heineman becky@burgerbecky.com

It is released under an MIT Open Source license. Please see LICENSE
for license details. Yes, you can use it in a
commercial title without paying anything, just give me a credit.
Please? It's not like I'm asking you for money!

"""

from __future__ import absolute_import, print_function, unicode_literals
import io
import os
import sys
import setuptools

# pylint: disable=import-outside-toplevel

CWD = os.path.dirname(os.path.abspath(__file__))

# Project specific strings
PROJECT_NAME = 'wslwinreg'
PROJECT_KEYWORDS = [
    'cygwin',
    'msys2',
    'wsl',
    'winreg',
    'development',
    'windows',
    'visual studio',
    'vs',
    'burger'
]

# Manually import the project
PROJECT_MODULE = __import__(PROJECT_NAME)

# Read me file is the long description
with io.open(os.path.join(CWD, 'README.rst'), encoding='utf-8') as filep:
    LONG_DESCRIPTION = filep.read()

# Create the dependency list
INSTALL_REQUIRES = [
    'setuptools >= 17.1',
    'enum34 >= 1.0.0'
]

# Project classifiers
CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Topic :: Software Development',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Natural Language :: English',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9'
]

# Extra files to include in the form of this tuple (directory,[files])
DATA_FILES = [
    # (PROJECT_NAME, ['LICENSE.txt'])
]

#
# Parms for setup
#

SETUP_ARGS = dict(
    name='wslwinreg',
    version=PROJECT_MODULE.__version__,

    # Use the readme as the long description
    description=PROJECT_MODULE.__summary__,
    long_description=LONG_DESCRIPTION,
    # long_description_content_type='text/x-rst; charset=UTF-8',
    license=PROJECT_MODULE.__license__,
    url=PROJECT_MODULE.__uri__,

    author=PROJECT_MODULE.__author__.encode('utf8'),
    author_email=PROJECT_MODULE.__email__,

    keywords=PROJECT_KEYWORDS,
    platforms=['Any'],
    install_requires=INSTALL_REQUIRES,
    zip_safe=False,
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',

    classifiers=CLASSIFIERS,
    packages=[PROJECT_NAME],
    include_package_data=True,
    data_files=DATA_FILES,
    package_data={
        PROJECT_NAME: ['bin/*.exe']
    }
)

########################################

CLEAN_DIR_LIST = [
    PROJECT_NAME + '.egg-info',
    PROJECT_NAME + '-' + PROJECT_MODULE.__version__,
    'dist',
    'build',
    'temp',
    '.pytest_cache',
    '.tox',
    '.vscode'
]

CLEAN_DIR_RECURSE_LIST = [
    'temp',
    '__pycache__',
    '_build',
    'bin',
    '.vs'
]

CLEAN_EXTENSION_LIST = [
    '*.pyc',
    '*.pyo',
    '*.user'
]


def clean(working_dir):
    """
    Clean up all the temp files after uploading

    Helps in keeping source control from having to track
    temp files
    """

    # Delete all folders, including read only files
    try:
        from burger import delete_directory, clean_directories, clean_files
    except ImportError as exc:
        print("Install python module burger with pip to run clean")
        raise exc

    for item in CLEAN_DIR_LIST:
        delete_directory(os.path.join(working_dir, item))

    clean_directories(
        working_dir,
        CLEAN_DIR_RECURSE_LIST,
        recursive=True)

    #
    # Delete all *.pyc and *.pyo files
    #

    clean_files(
        working_dir,
        name_list=CLEAN_EXTENSION_LIST,
        recursive=True)

#
# Perform the setup
#


if __name__ == '__main__':
    # Ensure the directory is the current one
    if CWD:
        os.chdir(CWD)

    # Perform a thorough cleaning job
    if 'clean' in sys.argv:
        clean(CWD)

    # Unlock the files to handle Perforce locking
    try:
        from burger import unlock_files, lock_files
    except ImportError:
        def lock_files(not_used):
            """ Stub """
            return not_used

        def unlock_files(not_used):
            """ Stub """
            return not_used

    LOCK_LIST = unlock_files(CWD) + \
        unlock_files(os.path.join(CWD, PROJECT_NAME))
    try:
        setuptools.setup(**SETUP_ARGS)

    # If any files were unlocked, relock them
    finally:
        lock_files(LOCK_LIST)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A drop in replacement for winreg for Cygwin, MSYS2 and Windows Subsystem for
Linux.
"""

## \package wslwinreg
#
# This module provides access to the Windows registry API.
#

#
## \mainpage
#
# \htmlinclude README.html
#
# Chapter list
# ============
#
# - \subpage md_why_wslwinreg
#
# Module list
# ===========
#
# - \ref wslwinreg
# - \ref wslwinreg.common
# - \ref wslwinreg.cygwinapi
# - \ref wslwinreg.wslapi
# - \ref wslwinreg.nullapi
# - \ref wslwinreg.WinRegKey
#

import itertools
import os.path

# pylint: disable=useless-object-inheritance
# pylint: disable=invalid-name

from .common import IS_CYGWIN, IS_MSYS, IS_WSL, ERROR_SUCCESS, \
    ERROR_FILE_NOT_FOUND, ERROR_MORE_DATA, HKEY_CLASSES_ROOT, \
    HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE, HKEY_USERS, HKEY_PERFORMANCE_DATA, \
    HKEY_CURRENT_CONFIG, HKEY_DYN_DATA, KEY_QUERY_VALUE, KEY_SET_VALUE, \
    KEY_CREATE_SUB_KEY, KEY_ENUMERATE_SUB_KEYS, KEY_NOTIFY, KEY_CREATE_LINK, \
    KEY_WOW64_32KEY, KEY_WOW64_64KEY, KEY_WOW64_RES, KEY_WRITE, KEY_EXECUTE, \
    KEY_READ, KEY_ALL_ACCESS, REG_OPTION_RESERVED, REG_OPTION_NON_VOLATILE, \
    REG_OPTION_VOLATILE, REG_OPTION_CREATE_LINK, REG_OPTION_BACKUP_RESTORE, \
    REG_OPTION_OPEN_LINK, REG_LEGAL_OPTION, REG_CREATED_NEW_KEY, \
    REG_OPENED_EXISTING_KEY, REG_WHOLE_HIVE_VOLATILE, REG_REFRESH_HIVE, \
    REG_NO_LAZY_FLUSH, REG_NOTIFY_CHANGE_NAME, REG_NOTIFY_CHANGE_ATTRIBUTES, \
    REG_NOTIFY_CHANGE_LAST_SET, REG_NOTIFY_CHANGE_SECURITY, \
    REG_LEGAL_CHANGE_FILTER, REG_NONE, REG_SZ, REG_EXPAND_SZ, REG_BINARY, \
    REG_DWORD, REG_DWORD_LITTLE_ENDIAN, REG_DWORD_BIG_ENDIAN, REG_LINK, \
    REG_MULTI_SZ, REG_RESOURCE_LIST, REG_FULL_RESOURCE_DESCRIPTOR, \
    REG_RESOURCE_REQUIREMENTS_LIST, REG_QWORD, REG_QWORD_LITTLE_ENDIAN, \
    FORMAT_MESSAGE_ALLOCATE_BUFFER, FORMAT_MESSAGE_IGNORE_INSERTS, \
    FORMAT_MESSAGE_FROM_STRING, FORMAT_MESSAGE_FROM_HMODULE, \
    FORMAT_MESSAGE_FROM_SYSTEM, FORMAT_MESSAGE_ARGUMENT_ARRAY, \
    FORMAT_MESSAGE_MAX_WIDTH_MASK, LANG_NEUTRAL, LPCVOID, BOOL, WORD, DWORD, \
    PDWORD, LPDWORD, QWORD, PQWORD, LPQWORD, LONG, PLONG, PBYTE, LPBYTE, \
    LPSTR, LPWSTR, LPCWSTR, HANDLE, HKEY, PHKEY, HLOCAL, REGSAM, FILETIME, \
    PFILETIME, SUBLANG_DEFAULT

## Numeric version
__numversion__ = (1, 0, 4)

## Current version of the library
__version__ = '.'.join([str(num) for num in __numversion__])

## Author's name
__author__ = 'Rebecca Ann Heineman'

## Name of the module
__title__ = 'wslwinreg'

## Summary of the module's use
__summary__ = 'Drop in replacement for winreg for Cygwin, MSYS2 and WSL'

## Home page
__uri__ = 'http://wslwinreg.readthedocs.io'

## Email address for bug reports
__email__ = 'becky@burgerbecky.com'

## Type of license used for distribution
__license__ = 'MIT License'

## Copyright owner
__copyright__ = 'Copyright 2020-2021 Rebecca Ann Heineman'

# Load in the proper implementation based on the
# underlying operating system

if IS_CYGWIN or IS_MSYS:
    from .cygwinapi import CloseKey, ConnectRegistry, CreateKey, CreateKeyEx, \
        DeleteKey, DeleteKeyEx, DeleteValue, EnumKey, EnumValue, \
        ExpandEnvironmentStrings, FlushKey, LoadKey, OpenKey, OpenKeyEx, \
        QueryInfoKey, QueryValue, QueryValueEx, SaveKey, SetValue, SetValueEx, \
        DisableReflectionKey, EnableReflectionKey, QueryReflectionKey, \
        get_file_info, convert_to_windows_path, convert_from_windows_path
elif IS_WSL:
    from .wslapi import CloseKey, ConnectRegistry, CreateKey, CreateKeyEx, \
        DeleteKey, DeleteKeyEx, DeleteValue, EnumKey, EnumValue, \
        ExpandEnvironmentStrings, FlushKey, LoadKey, OpenKey, OpenKeyEx, \
        QueryInfoKey, QueryValue, QueryValueEx, SaveKey, SetValue, SetValueEx, \
        DisableReflectionKey, EnableReflectionKey, QueryReflectionKey, \
        get_file_info, convert_to_windows_path, convert_from_windows_path
else:
    from .nullapi import convert_to_windows_path, convert_from_windows_path
    try:
        # Attempt importing the current name
        from winreg import *
        from .cygwinapi import get_file_info
    except ImportError:
        try:
            # Attempt importing the old name
            from _winreg import *
            from .cygwinapi import get_file_info
        except ImportError:
            # For unsupported platforms, create null apis that always
            # throw exceptions when called
            from .nullapi import CloseKey, ConnectRegistry, CreateKey, \
                CreateKeyEx, DeleteKey, DeleteKeyEx, DeleteValue, EnumKey, \
                EnumValue, ExpandEnvironmentStrings, FlushKey, LoadKey, \
                OpenKey, OpenKeyEx, QueryInfoKey, QueryValue, QueryValueEx, \
                SaveKey, SetValue, SetValueEx, DisableReflectionKey, \
                EnableReflectionKey, QueryReflectionKey, get_file_info

########################################


class WinRegKey(object):
    """
    Registry key helper class.

    Manage entries in a registy key by using indexing and scanning.
    """

    def __init__(self, root_key, subkey, access):
        """
        Initialize the class

        Args:
            root_key: Root key enumeration
            subkey: String or None for name of sub key
            access: Access flags to pass to OpenKeyEx()
        """

        # If there is a sub key, open it, can throw an exception
        if subkey:
            root_key = OpenKeyEx(root_key, subkey, 0, access)

        ## Key handle of opened key
        self.key = root_key

        ## Access used to open this key, used as default to open sub keys
        self.access = access

    def __enter__(self):
        """
        Enable enter/exit functionality
        """
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """
        Release resources on class release

        Args:
            exception_type: Ignored
            exception_value: Ignored
            traceback: Ignored
        """
        self.close()

    def close(self):
        """
        Release existing key, if any.
        """
        CloseKey(self.key)

    def open_subkey(self, subkey, access=None):
        """
        Open a sub key.

        Args:
            subkey: Name of the subkey to open
            access: Desired access to the subkey.
        Returns:
            An instance of the WinRegKey of the new key.
        """

        # Use the default access?
        if access is None:
            access = self.access
        return type(self)(self.key, subkey, access)

    def get_subkeys(self):
        """
        Return the list of all of the sub key names.

        Iterate over all of the sub keys and place their names in a list.
        Return the list. Unicode is properly handled.

        Returns:
            list of names of all the subkeys.
        """
        subkey_names = []
        try:
            for item in itertools.count():
                subkey_names.append(EnumKey(self.key, item))

        # Exception is fired once the list end is reached
        except OSError:
            pass
        return subkey_names

    def get_value(self, value_name=None):
        """
        Read a value from a registry key.

        Note:
            If the type returned is REG_SZ, scan for a null and trucate.
            If the type returned is REG_EXPAND_SZ, call
            ExpandEnvironmentStrings()

        Args:
            value_name: String of the name of the value, None for default.
        Returns:
            Value returned from QueryValueEx()
        """
        value, value_type = QueryValueEx(self.key, value_name)
        if value_type == REG_SZ:
            # Failsafe, truncate at null
            index = value.find('\0')
            if index != -1:
                value = value[:index]
        elif value_type == REG_EXPAND_SZ:
            # Perform expansion
            value = ExpandEnvironmentStrings(value)
        return (value, value_type)

    def get_all_values(self):
        """
        Return a dict of all key name and associated values.

        Returns:
            dict with each item is the returned value from QueryValueEx()
        """
        value_names = []
        try:
            for i in itertools.count():
                value_names.append(EnumValue(self.key, i))

        # Exception is fired once the list end is reached
        except OSError:
            pass

        # Convert the list to a dict
        return {k[0]: self.get_value(k[0]) for k in value_names}

    def __getitem__(self, subkey):
        """
        Call open_subkey() with subscript.

        Uses the access flags from this key to open the sub key.

        Args:
            subkey: Requested sub key.
        Returns:
            An instance of the WinRegKey of the new key.
        """
        return self.open_subkey(subkey)

    def __iter__(self):
        """
        Iterator of the sub keys.

        Call get_subkeys() and encapsulate the value in an iter()

        Returns:
            iter of get_subkeys()
        """
        return iter(self.get_subkeys())


########################################

def get_HKCU():
    """
    Open HKEY_CURRENT_USER.

    Returns:
        Read only WinRegKey for HKEY_CURRENT_USER
    """
    return WinRegKey(HKEY_CURRENT_USER, None, KEY_READ)

########################################


def get_HKLM_32():
    """
    Open HKEY_LOCAL_MACHINE, 32 bit view.

    Returns:
        Read only WinRegKey for HKEY_LOCAL_MACHINE with KEY_WOW64_32KEY
    """
    return WinRegKey(HKEY_LOCAL_MACHINE, None, KEY_READ | KEY_WOW64_32KEY)

########################################


def get_HKLM_64():
    """
    Open HKEY_LOCAL_MACHINE, 64 bit view.

    Returns:
        Read only WinRegKey for HKEY_LOCAL_MACHINE with KEY_WOW64_64KEY
    """
    return WinRegKey(HKEY_LOCAL_MACHINE, None, KEY_READ | KEY_WOW64_64KEY)
